import networkx as nx
import numpy as np
import pandas as pd
import pytest

import treedata as td


@pytest.fixture
def tree():
    tree = nx.balanced_tree(r=2, h=3, create_using=nx.DiGraph)
    tree = nx.relabel_nodes(tree, {i: str(i) for i in tree.nodes})
    depths = nx.single_source_shortest_path_length(tree, "0")
    nx.set_node_attributes(tree, values=depths, name="depth")
    yield tree


@pytest.fixture
def tdata(tree):
    df = pd.DataFrame({"anno": range(8)}, index=[str(i) for i in range(7, 15)])
    yield td.TreeData(X=np.zeros((8, 8)), obst={"tree": tree}, vart={"tree": tree}, obs=df, var=df, allow_overlap=True)


@pytest.fixture
def nodes_tdata(tree):
    df = pd.DataFrame({"anno": range(16)}, index=[str(i) for i in range(16)])
    yield td.TreeData(X=np.zeros((16, 16)), obst={"tree": tree}, vart={"tree": tree}, obs=df, var=df, alignment="nodes")


def test_views(tdata):
    # check that subset is view
    assert tdata[:, 0].is_view
    assert tdata[:2, 0].X.shape == (2, 1)
    tdata_subset = tdata[:2, [0, 1]]
    assert tdata_subset.is_view
    # now transition to actual object
    with pytest.warns(UserWarning):
        tdata_subset.obs["test"] = range(2)
    assert not tdata_subset.is_view
    assert tdata_subset.obs["test"].tolist() == list(range(2))


def test_views_subset_leaves(tdata):
    expected_edges = [
        ("0", "1"),
        ("0", "2"),
        ("1", "3"),
        ("2", "5"),
        ("3", "7"),
        ("3", "8"),
        ("5", "11"),
    ]
    # subset with index
    tdata_subset = tdata[[0, 1, 4], :]
    edges = list(tdata_subset.obst["tree"].edges)
    assert edges == expected_edges
    # subset with names
    tdata_subset = tdata[["7", "8", "11"], :]
    edges = list(tdata_subset.obst["tree"].edges)
    assert edges == expected_edges
    # now transition to actual object
    tdata_subset = tdata_subset.copy()
    edges = list(tdata_subset.obst["tree"].edges)
    assert edges == expected_edges
    assert len(tdata.obst["tree"].edges) == 14
    # now transition alignment to subset
    tdata.alignment = "subset"
    tdata_subset = tdata[[0, 1, 4], :]
    assert tdata_subset.alignment == "subset"
    assert tdata_subset.shape == (3, 8)
    assert len(tdata_subset.obst["tree"].nodes) == 15


def test_views_subset_nodes(nodes_tdata):
    expected_edges = [
        ("0", "1"),
        ("0", "2"),
        ("1", "3"),
    ]
    # subset with index
    tdata_subset = nodes_tdata[[0, 1, 2, 3], :]
    edges = list(tdata_subset.obst["tree"].edges)
    assert set(edges) == set(expected_edges)
    # subset with names
    tdata_subset = nodes_tdata[["0", "1", "2", "3"], :]
    edges = list(tdata_subset.obst["tree"].edges)
    assert set(edges) == set(expected_edges)
    # now transition to actual object
    tdata_subset = tdata_subset.copy()
    edges = list(tdata_subset.obst["tree"].edges)
    assert set(edges) == set(expected_edges)
    assert len(nodes_tdata.obst["tree"].edges) == 14
    # invalid tree switches alignment to subset
    with pytest.warns(UserWarning):
        tdata_subset = nodes_tdata[["0", "1", "5"], :]
    assert tdata_subset.alignment == "subset"
    assert not tdata_subset.is_view
    assert len(tdata_subset.obst["tree"].nodes) == 15
    assert len(tdata_subset.vart["tree"].edges) == 14
    assert tdata_subset.shape == (3, 16)


def test_views_subset_trees():
    # subset tdata with multiple trees
    tree1 = nx.DiGraph([("root", "0"), ("root", "1")])
    tree2 = nx.DiGraph([("root", "2"), ("root", "3")])
    tdata = td.TreeData(X=np.zeros((8, 8)), allow_overlap=False, obst={"tree1": tree1, "tree2": tree2})
    tdata_subset = tdata[["0", "1"], :]
    assert list(tdata_subset.obst["tree1"].edges) == [("root", "0"), ("root", "1")]
    assert list(tdata_subset.obst["tree2"].edges) == []
    tdata_subset = tdata[["2", "3"], :]
    assert list(tdata_subset.obst["tree1"].edges) == []
    assert list(tdata_subset.obst["tree2"].edges) == [("root", "2"), ("root", "3")]
    tdata_subset = tdata[["0", "1", "2"], :]
    assert list(tdata_subset.obst["tree1"].edges) == [("root", "0"), ("root", "1")]
    assert list(tdata_subset.obst["tree2"].edges) == [("root", "2")]
    # node subset with multiple trees
    obs = pd.DataFrame(index=["root", "0", "1", "2", "3"])
    tdata = td.TreeData(obs=obs, allow_overlap=True, obst={"tree1": tree1, "tree2": tree2})
    tdata_subset = tdata[["root", "0", "1"], :]
    assert list(tdata_subset.obst["tree1"].edges) == [("root", "0"), ("root", "1")]
    assert list(tdata_subset.obst["tree2"].edges) == []
    assert list(tdata_subset.obst.keys()) == ["tree1", "tree2"]


def test_views_copy():
    # subset tdata with multiple trees
    tree1 = nx.DiGraph([("root", "0"), ("root", "1")])
    tree2 = nx.DiGraph([("root", "2"), ("root", "3")])
    tdata = td.TreeData(X=np.zeros((8, 8)), allow_overlap=False, obst={"tree1": tree1, "tree2": tree2})
    tdata_subset = tdata.copy()[["0", "1"], :].copy()
    assert list(tdata_subset.obst["tree1"].edges) == [("root", "0"), ("root", "1")]
    assert list(tdata_subset.obst["tree2"].edges) == []
    tdata_subset = tdata.copy()[["2", "3"], :].copy()
    assert list(tdata_subset.obst["tree1"].edges) == []
    assert list(tdata_subset.obst["tree2"].edges) == [("root", "2"), ("root", "3")]
    tdata_subset = tdata.copy()[["0", "1", "2"], :].copy()
    assert list(tdata_subset.obst["tree1"].edges) == [("root", "0"), ("root", "1")]
    assert list(tdata_subset.obst["tree2"].edges) == [("root", "2")]


def test_views_mutability(tdata):
    # can mutate attributes of graph
    nx.set_node_attributes(tdata.obst["tree"], False, "in_subset")
    subset_leaves = ["7", "8"]
    tdata_subset = tdata[subset_leaves, :]
    nx.set_node_attributes(tdata_subset.obst["tree"], True, "in_subset")
    expected_subset_nodes = ["8", "0", "3", "7", "1"]
    subset_nodes = [
        node for node in tdata_subset.obst["tree"].nodes if tdata_subset.obst["tree"].nodes[node]["in_subset"]
    ]
    assert set(subset_nodes) == set(expected_subset_nodes)
    # cannot mutate structure of graph
    with pytest.raises(nx.NetworkXError):
        tdata_subset.obst["tree"].remove_node("8")


def test_views_set(tdata):
    tdata_subset = tdata[[0, 1, 4], :]
    # bad assignment
    bad_tree = nx.DiGraph()
    bad_tree.add_edge("0", "bad")
    with pytest.raises(ValueError):
        tdata_subset.obst["new_tree"] = bad_tree
    assert tdata_subset.is_view
    # good assignment actualizes object
    new_tree = nx.DiGraph()
    new_tree.add_edge("0", "8")
    with pytest.warns(UserWarning):
        tdata_subset.obst["new_tree"] = new_tree
    assert not tdata_subset.is_view
    assert list(tdata_subset.obst.keys()) == ["tree", "new_tree"]
    assert list(tdata_subset.obst["new_tree"].edges) == [("0", "8")]
    # good assignment no overlap
    tree = nx.DiGraph([("root", "0"), ("root", "1")])
    good_tree = nx.DiGraph([("root", "2"), ("root", "3")])
    tdata = td.TreeData(X=np.zeros((8, 8)), allow_overlap=False, obst={"tree": tree})
    tdata_subset = tdata[:4, :]
    with pytest.warns(UserWarning):
        tdata_subset.obst["tree"] = good_tree


def test_views_del(tdata):
    tdata_subset = tdata[[0, 1, 4], :]
    # bad deletion
    with pytest.raises(KeyError):
        del tdata_subset.obst["bad"]
    assert tdata_subset.is_view
    # good deletion actualizes object
    with pytest.warns(UserWarning):
        del tdata_subset.obst["tree"]
    assert not tdata_subset.is_view
    assert list(tdata_subset.obst.keys()) == []


def test_views_contains(tdata):
    tdata_subset = tdata[[0, 1, 4], :]
    assert "tree" in tdata_subset.obst
    assert "bad" not in tdata_subset.obst


def test_views_len(tdata):
    tdata_subset = tdata[[0, 1, 4], :]
    assert len(tdata_subset.obst) == 1


if __name__ == "__main__":
    pytest.main(["-v", __file__])



import anndata as ad
import joblib
import networkx as nx
import numpy as np
import pandas as pd
import pytest

import treedata as td


@pytest.fixture
def X():
    yield np.array([[1, 2, 3], [4, 5, 6], [7, 8, 9]])


@pytest.fixture
def tree():
    tree = nx.DiGraph()
    tree.add_edges_from([("root", "0"), ("root", "1")])
    tree["root"]["0"]["length"] = 1
    tree.nodes["root"]["depth"] = 0
    tree.nodes["root"]["characters"] = ["-1", "1"]
    yield tree


@pytest.fixture
def tdata(X, tree):
    yield td.TreeData(X, obst={"1": tree, "2": tree}, vart={"1": tree}, label="tree", allow_overlap=True)


def check_graph_equality(g1, g2):
    assert nx.is_isomorphic(
        g1, g2, node_match=lambda n1, n2: set(n1) == set(n2), edge_match=lambda e1, e2: set(e1) == set(e2)
    )


@pytest.mark.parametrize("backed", [None, "r"])
def test_h5td_readwrite(tdata, tmp_path, backed):
    tdata.raw = tdata
    file_path = tmp_path / "test.h5td"
    tdata.write_h5td(file_path)
    tdata2 = td.read_h5td(file_path, backed=backed)
    assert np.array_equal(tdata2.X, tdata.X)
    assert np.array_equal(tdata2.raw.X, tdata.raw.X)
    check_graph_equality(tdata2.obst["1"], tdata.obst["1"])
    check_graph_equality(tdata2.vart["1"], tdata.vart["1"])
    assert tdata2.label == "tree"
    assert tdata2.allow_overlap is True
    assert tdata2.alignment == "leaves"
    assert tdata2.obst["1"].nodes["root"]["depth"] == 0
    assert tdata2.obst["2"].nodes["root"]["characters"] == ["-1", "1"]
    assert tdata2.obs.loc["0", "tree"] == "1,2"
    if backed:
        assert tdata2.isbacked
        assert tdata2.file.is_open
        assert tdata2.filename == file_path


def test_h5td_dtypes(tdata, tmp_path):
    file_path = tmp_path / "test.h5td"
    tdata.obst["1"].nodes["root"]["list"] = [1, 2, 3]
    tdata.obst["1"].nodes["root"]["tuple"] = (1, 2, 3)
    tdata.obst["1"].nodes["root"]["set"] = {1, 2, 3}
    tdata.obst["1"].nodes["root"]["np_float"] = np.float64(1.0)
    tdata.obst["1"].nodes["root"]["np_array"] = np.array([[1, 2], [3, 4]])
    tdata.obst["1"].nodes["root"]["pd_series"] = pd.Series(["1", "2", "3"])
    tdata.write_h5td(file_path)
    tdata2 = td.read_h5td(file_path)
    assert tdata2.obst["1"].nodes["root"]["list"] == [1, 2, 3]
    assert isinstance(tdata2.obst["1"].nodes["root"]["list"], list)
    assert tdata2.obst["1"].nodes["root"]["tuple"] == [1, 2, 3]
    assert isinstance(tdata2.obst["1"].nodes["root"]["tuple"], list)
    assert tdata2.obst["1"].nodes["root"]["set"] == [1, 2, 3]
    assert isinstance(tdata2.obst["1"].nodes["root"]["set"], list)
    assert tdata2.obst["1"].nodes["root"]["np_float"] == 1.0
    assert isinstance(tdata2.obst["1"].nodes["root"]["np_float"], float)
    assert tdata2.obst["1"].nodes["root"]["np_array"] == [[1, 2], [3, 4]]
    assert isinstance(tdata2.obst["1"].nodes["root"]["np_array"], list)
    assert tdata2.obst["1"].nodes["root"]["pd_series"] == ["1", "2", "3"]
    assert isinstance(tdata2.obst["1"].nodes["root"]["pd_series"], list)


def test_zarr_readwrite(tdata, tmp_path):
    tdata.write_zarr(tmp_path / "test.zarr")
    tdata2 = td.read_zarr(tmp_path / "test.zarr")
    assert np.array_equal(tdata2.X, tdata.X)
    check_graph_equality(tdata2.obst["1"], tdata.obst["1"])
    check_graph_equality(tdata2.vart["1"], tdata.vart["1"])
    assert tdata2.label == "tree"
    assert tdata2.allow_overlap is True
    assert tdata2.alignment == "leaves"
    assert tdata2.obst["2"].nodes["root"]["depth"] == 0
    assert tdata2.obs.loc["0", "tree"] == "1,2"


def test_read_anndata(X, tmp_path):
    adata = ad.AnnData(X)
    file_path = tmp_path / "test.h5ad"
    adata.write_h5ad(file_path)
    tdata = td.read_h5td(file_path)
    assert np.array_equal(tdata.X, adata.X)
    assert tdata.label == "tree"
    assert tdata.allow_overlap is True
    assert tdata.alignment == "leaves"
    assert list(tdata.obst.keys()) == []


def test_read_no_X(X, tmp_path):
    tdata = td.TreeData(obs=pd.DataFrame(index=["0", "1", "2"]))
    file_path = tmp_path / "test.h5td"
    tdata.write_h5td(file_path)
    tdata2 = td.read_h5td(file_path)
    assert tdata2.X is None


def deprecated_read_write(tdata, tmp_path):
    # Test deprecated read/write methods
    file_path = tmp_path / "test_deprecated.h5td"
    with pytest.warns(DeprecationWarning):
        tdata.write_h5ad(file_path)
    with pytest.warns(DeprecationWarning):
        td.read_h5ad(file_path)


def test_h5td_backing(tdata, tree, tmp_path):
    tdata_copy = tdata.copy()
    assert not tdata.isbacked
    backing_h5td = tmp_path / "test_backed.h5td"
    tdata.filename = backing_h5td
    # backing mode
    tdata.write()
    assert not tdata.file.is_open
    assert tdata.isbacked
    # view of backed object
    tdata_subset = tdata[:, 0]
    subset_hash = joblib.hash(tdata_subset)
    assert tdata_subset.is_view
    assert tdata_subset.isbacked
    assert tdata_subset.shape == (3, 1)
    check_graph_equality(tdata_subset.obst["1"], tdata.obst["1"])
    assert np.array_equal(tdata_subset.X, tdata_copy.X[:, 0].reshape(-1, 1))
    # cannot set view in backing mode...
    with pytest.warns(UserWarning):
        with pytest.raises(ValueError):
            tdata_subset.obs["foo"] = range(3)
    assert subset_hash == joblib.hash(tdata_subset)
    assert tdata_subset.is_view
    # copy
    tdata_subset = tdata_subset.copy(tmp_path / "test_subset.h5td")
    assert not tdata_subset.is_view
    tdata_subset.obs["foo"] = range(3)
    assert not tdata_subset.is_view
    assert tdata_subset.isbacked
    assert tdata_subset.obs["foo"].tolist() == list(range(3))
    tdata_subset.write()
    # move to memory
    tdata_subset = tdata_subset.to_memory()
    assert not tdata_subset.is_view
    assert not tdata_subset.isbacked
    print(tdata_subset)
    check_graph_equality(tdata_subset.obst["1"], tdata.obst["1"])


if __name__ == "__main__":
    pytest.main(["-v", __file__])



from __future__ import annotations

import json
import warnings
from collections.abc import MutableMapping
from importlib.metadata import version as get_version
from typing import TYPE_CHECKING, Literal

import anndata as ad
import h5py
import networkx as nx
import zarr
from packaging import version

from treedata._core.treedata import TreeData

if TYPE_CHECKING:
    from os import PathLike

ANNDATA_VERSION = version.parse(get_version("anndata"))
USE_EXPERIMENTAL = ANNDATA_VERSION < version.parse("0.11.0")


def _read_elem(elem):
    """Read an element from a store."""
    if USE_EXPERIMENTAL:
        return ad.experimental.read_elem(elem)
    else:
        return ad.io.read_elem(elem)


def _dict_to_digraph(graph_dict: dict) -> nx.DiGraph:
    """Convert a dictionary to a networkx.DiGraph."""
    G = nx.DiGraph()
    # Add nodes and their attributes
    for node, attrs in graph_dict["nodes"].items():
        G.add_node(node, **attrs)
    # Add edges and their attributes
    for source, targets in graph_dict["edges"].items():
        for target, attrs in targets.items():
            G.add_edge(source, target, **attrs)
    return G


def _parse_axis_trees(data: str) -> dict:
    """Parse AxisTrees from a string."""
    return {k: _dict_to_digraph(v) for k, v in json.loads(data).items()}


def _parse_legacy(treedata_attrs: dict) -> dict:
    """Parse tree attributes from AnnData uns field."""
    if treedata_attrs is not None:
        for j in ["obst", "vart"]:
            if j in treedata_attrs:
                treedata_attrs[j] = {k: _dict_to_digraph(v) for k, v in treedata_attrs[j].items()}
        treedata_attrs["allow_overlap"] = bool(treedata_attrs["allow_overlap"])
        treedata_attrs["label"] = treedata_attrs["label"] if "label" in treedata_attrs.keys() else None
    return treedata_attrs


def _read_raw(f, backed):
    """Read raw from file."""
    d = {}
    for k in ["obs", "var"]:
        if f"raw/{k}" in f:
            d[k] = _read_elem(f[f"raw/{k}"])
    if not backed:
        d["X"] = _read_elem(f["raw/X"])
    return d


def _read_tdata(f, filename, backed) -> dict:
    """Read TreeData from file."""
    d = {}
    if backed is None:
        backed = False
    elif backed is True:
        backed = "r"
    # Read X if not backed
    if not backed:
        if "X" in f:
            d["X"] = _read_elem(f["X"])
    else:
        d.update({"filename": filename, "filemode": backed})
    # Read standard elements
    for k in ["obs", "var", "obsm", "varm", "obsp", "varp", "layers", "uns", "label", "allow_overlap", "alignment"]:
        if k in f:
            d[k] = _read_elem(f[k])
    # Read raw
    if "raw" in f:
        d["raw"] = _read_raw(f, backed)
    # Read axis tree elements
    for k in ["obst", "vart"]:
        if k in f:
            d[k] = _parse_axis_trees(_read_elem(f[k]))
    # Read legacy treedata format
    if "raw.treedata" in f:
        d.update(_parse_legacy(json.loads(_read_elem(f["raw.treedata"]))))
    return d


def read_h5td(
    filename: str | PathLike | None = None,
    backed: Literal["r", "r+"] | bool | None = None,
) -> TreeData:
    """Read `.h5td` or `.h5ad`-formatted hdf5 file.

    Parameters
    ----------
    filename
        File name of data file.
    backed
        If `'r'`, load :class:`~TreeData` in `backed` mode
        instead of fully loading it into memory (`memory` mode).
        If you want to modify backed attributes of the TreeData object,
        you need to choose `'r+'`.
    """
    with h5py.File(filename, "r") as f:
        d = _read_tdata(f, filename, backed)
    return TreeData(**d)


def read_h5ad(
    filename: str | PathLike | None = None,
    backed: Literal["r", "r+"] | bool | None = None,
) -> TreeData:
    """Read `.h5td` or `.h5ad`-formatted hdf5 file. Deprecated, use `read_h5td` instead.

    Parameters
    ----------
    filename
        File name of data file.
    backed
        If `'r'`, load :class:`~TreeData` in `backed` mode
        instead of fully loading it into memory (`memory` mode).
        If you want to modify backed attributes of the TreeData object,
        you need to choose `'r+'`.
    """
    warnings.warn(
        "read_h5ad has been renamed to read_h5td. read_h5ad will be removed in v1.0.0.",
        DeprecationWarning,
        stacklevel=2,
    )
    return read_h5td(filename, backed=backed)


def _open_zarr_group(store, *, mode: str = "r") -> tuple[zarr.Group, bool]:
    """Open a Zarr group and signal whether it should be closed."""
    if isinstance(store, zarr.Group):
        return store, False
    return zarr.open(store, mode=mode), True


def _close_zarr_group(group: zarr.Group) -> None:
    """Close the underlying store for an opened Zarr group if needed."""
    store = getattr(group, "store", None)
    if store is not None:
        close = getattr(store, "close", None)
        if callable(close):
            close()


def read_zarr(store: str | PathLike | MutableMapping | zarr.Group) -> TreeData:
    """Read from a hierarchical Zarr array store.

    Parameters
    ----------
    store
        The filename, a :class:`~typing.MutableMapping`, or a Zarr storage class.
    """
    group, should_close = _open_zarr_group(store)
    d = _read_tdata(group, store, backed=False)
    if should_close:
        _close_zarr_group(group)
    return TreeData(**d)
